/**********************************************************************
 *<
	FILE: afregion.cpp

	DESCRIPTION:  Affect region modifier

	CREATED BY: Rolf Berteig

	HISTORY: 10/16/96

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

#include "mods.h"
#include "meshadj.h"
#include "iparamm.h"

// This is based on the simple spline object...

#include "Simpobj.h"
#include "Metapart.h"

class NullView: public View {
	public:
		Point2 ViewToScreen(Point3 p) { return Point2(p.x,p.y); }
		NullView() { worldToView.IdentityMatrix(); screenW=640.0f; screenH = 480.0f; }
	};


#define A_RENDER			A_PLUGIN1

#define NODE_COUNT_CHUNK		0x10

#define PBLOCK_REF	0
#define NODE_REF	1
#define LAZYID 0xDE17A356, 0x8A41f3aa

class AFRMod : public Modifier {	
	public:
		IParamBlock *pblock;
		HWND hWnd;
		HWND ListhWnd;
		static IObjParam *ip;
		static IParamMap *pmapParam;
		ICustButton *iBut;// = GetICustButton(GetDlgItem(mod->hParams,IDC_GRAVITY_PICKNODE));
		BlobParticle blobP;		// blob object used



		INode *SelfNode;     //node to base object
		Tab<INode*> OtherNode;
		void RemoveNode(int index);

		int RenderBegin(TimeValue t, ULONG flags);

		int RenderEnd(TimeValue t);

		
		AFRMod();
		~AFRMod() {
				DeleteAllRefsFromMe();
				pblock = NULL;
				if (iBut)
					ReleaseICustButton(iBut);
				iBut = NULL;

				}

		float size,tension, coursenessR,coursenessV;
		int autoCourse,hideEdges;

		// From Animatable
		void DeleteThis() { delete this; }
		void GetClassName(TSTR& s) {s = GetString(IDS_RB_AFRMOD);}  
		virtual Class_ID ClassID() { return Class_ID(LAZYID);}		
		void BeginEditParams(IObjParam  *ip, ULONG flags,Animatable *prev);
		void EndEditParams(IObjParam *ip,ULONG flags,Animatable *next);		
		RefTargetHandle Clone(RemapDir& remap = NoRemap());
		TCHAR *GetObjectName() {return GetString(IDS_RB_AFRMOD);}

		// From modifier
		// From modifier
		ChannelMask ChannelsUsed()  {return PART_GEOM|PART_TOPO|PART_TEXMAP|PART_SELECT|PART_SUBSEL_TYPE|PART_VERTCOLOR;}
		ChannelMask ChannelsChanged() {return PART_GEOM|PART_TOPO|PART_SELECT|PART_TEXMAP|PART_VERTCOLOR;}
		Class_ID InputType() {return triObjectClassID;}
	
		int Display(TimeValue t, INode* inode, ViewExp *vpt, int flagst, ModContext *mc);

		void ModifyObject(TimeValue t, ModContext &mc, ObjectState *os, INode *node);
		Interval LocalValidity(TimeValue t);

		// From BaseObject
		CreateMouseCallBack* GetCreateMouseCallBack() {return NULL;} 

		int NumRefs() {return 1+OtherNode.Count();}
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);

		int NumSubs() {return 1;}
		Animatable* SubAnim(int i); 
		TSTR SubAnimName(int i);

		RefResult NotifyRefChanged( Interval changeInt,RefTargetHandle hTarget, 
		   PartID& partID, RefMessage message);

        IOResult Load(ILoad *iload);
        IOResult Save(ISave *isave);

	};




//--- ClassDescriptor and class vars ---------------------------------

IParamMap       *AFRMod::pmapParam = NULL;
IObjParam       *AFRMod::ip        = NULL;
		

class AFRClassDesc:public ClassDesc {
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading = FALSE) { return new AFRMod; }
	const TCHAR *	ClassName() { return GetString(IDS_RB_AFRMOD); }
	SClass_ID		SuperClassID() { return OSM_CLASS_ID; }
	Class_ID		ClassID() { return Class_ID(LAZYID); }
	const TCHAR* 	Category() { return GetString(IDS_RB_DEFDEFORMATIONS);}
	};

static AFRClassDesc afrDesc;
extern ClassDesc* GetAFRModDesc() {return &afrDesc;}


//--- Parameter map/block descriptors -------------------------------


#define PB_SIZE			0
#define PB_TENSION		1
#define PB_RENDER		2
#define PB_VIEW			3
#define PB_AUTO			4
#define PB_HIDE			5


//
//
// Parameters

static ParamUIDesc descParam[] = {

	// Size
	ParamUIDesc(
		PB_SIZE,
		EDITTYPE_FLOAT,
		IDC_SIZE2,IDC_SIZE2SPIN,
		0.1f,10000.0f,
		SPIN_AUTOSCALE),

	// Tension
	ParamUIDesc(
		PB_TENSION,
		EDITTYPE_FLOAT,
		IDC_SP_METTENS,IDC_SP_METTENSSPIN,
		0.01f,10000.0f,
		SPIN_AUTOSCALE),

	// Render
	ParamUIDesc(
		PB_RENDER,
		EDITTYPE_FLOAT,
		IDC_SP_METCOURSE,IDC_SP_METCOURSESPIN,
		0.01f,10000.0f,
		SPIN_AUTOSCALE),

	// View
	ParamUIDesc(
		PB_VIEW,
		EDITTYPE_FLOAT,
		IDC_SP_METCOURSEV,IDC_SP_METCOURSEVSPIN,
		0.01f,10000.0f,
		SPIN_AUTOSCALE),

	// Auto
   ParamUIDesc(PB_AUTO,TYPE_SINGLECHEKBOX,IDC_SP_AUTOCOARSE),

	// Hide
   ParamUIDesc(PB_HIDE,TYPE_SINGLECHEKBOX,IDC_HIDE),



	};

#define PARAMDESC_LENGTH	6

static ParamBlockDescID descVer0[] = {
	{ TYPE_FLOAT, NULL, TRUE,  0 },		// Size	
	{ TYPE_FLOAT, NULL, TRUE,  1 },		// Tension
	{ TYPE_FLOAT, NULL, TRUE,  2 },		// Render	
	{ TYPE_FLOAT, NULL, TRUE,  3 },		// View
	{ TYPE_INT, NULL, FALSE,  4 },		// Auto
	{ TYPE_INT, NULL, FALSE,  5 },		// HIDE
	};
#define PBLOCK_LENGTH	6

#define CURRENT_VERSION	0



//--- CustMod dlg proc for the pick gravity------------------------------

class PickControlNode : 
		public PickModeCallback,
		public PickNodeCallback {
	public:				
		AFRMod *mod;
		PickControlNode() {mod=NULL;}
		BOOL HitTest(IObjParam *ip,HWND hWnd,ViewExp *vpt,IPoint2 m,int flags);		
		BOOL Pick(IObjParam *ip,ViewExp *vpt);		
		void EnterMode(IObjParam *ip);
		void ExitMode(IObjParam *ip);		
		BOOL Filter(INode *node);
		PickNodeCallback *GetFilter() {return this;}
		BOOL RightClick(IObjParam *ip,ViewExp *vpt) {return TRUE;}
	};
static PickControlNode thePickMode;

BOOL PickControlNode::Filter(INode *node)
	{
	node->BeginDependencyTest();
	mod->NotifyDependents(FOREVER,0,REFMSG_TEST_DEPENDENCY);
	if (node->EndDependencyTest()) 
		{		
		return FALSE;
		} 
	else 
		{
		for (int i = 0;i < mod->OtherNode.Count(); i++)
			{
			if (node == mod->OtherNode[i])
				return FALSE;
			}
		const ObjectState& os = node->EvalWorldState(0);
		Object* ob = os.obj;
		if (ob!=NULL) 
			{	
			return TRUE;
				  
			}
			else return FALSE;
		}
	return FALSE;
	
	}

BOOL PickControlNode::HitTest(
		IObjParam *ip,HWND hWnd,ViewExp *vpt,IPoint2 m,int flags)
	{	
	if (ip->PickNode(hWnd,m,this)) {
		return TRUE;
	} else {
		return FALSE;
		}
	}

BOOL PickControlNode::Pick(IObjParam *ip,ViewExp *vpt)
	{
	INode *node = vpt->GetClosestHit();
	if (node) {
		INode *tnode = NULL;
		mod->OtherNode.Append(1,&tnode,1);

		mod->ReplaceReference(mod->OtherNode.Count(),node);


		mod->NotifyDependents(FOREVER,0,REFMSG_CHANGE);

		TCHAR title[200];
		_tcscpy(title,node->GetName());
		SendMessage(mod->ListhWnd,LB_ADDSTRING,0,(LPARAM)(TCHAR*)title);



		return FALSE;
		}


	return FALSE;
	}

void PickControlNode::EnterMode(IObjParam *ip)
	{
	if (mod->iBut) mod->iBut->SetCheck(TRUE);
	}

void PickControlNode::ExitMode(IObjParam *ip)
	{
	if (mod->iBut) mod->iBut->SetCheck(FALSE);


	}


class MapDlgProc : public ParamMapUserDlgProc {
	public:
		AFRMod *mod;		
		MapDlgProc(AFRMod *m) {mod = m;}		
		BOOL DlgProc(TimeValue t,IParamMap *map,HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam);		
		void DeleteThis() {delete this;}
	};

BOOL MapDlgProc::DlgProc(
		TimeValue t,IParamMap *map,HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam)
	{
	int i;
	switch (msg) {
		case WM_INITDIALOG:


			mod->hWnd = hWnd;
			mod->iBut = GetICustButton(GetDlgItem(hWnd,IDC_ADD));
			mod->iBut->SetType(CBT_CHECK);
			mod->iBut->SetHighlightColor(GREEN_WASH);
			mod->ListhWnd = GetDlgItem(hWnd,IDC_LIST1);
			for (i = 0; i< mod->OtherNode.Count();i++)
				{
				if (mod->OtherNode[i] != NULL)
					{
					TCHAR title[200];
					_tcscpy(title,mod->OtherNode[i]->GetName());
					SendMessage(GetDlgItem(mod->hWnd,IDC_LIST1),
						LB_ADDSTRING,0,(LPARAM)(TCHAR*)title);
					}

				}


			break;

		case WM_COMMAND:

			switch (LOWORD(wParam)) {
			
				case IDC_ADD:
					{
					thePickMode.mod  = mod;					
					mod->ip->SetPickMode(&thePickMode);

					break;
					}
				case IDC_REMOVE:
					{
					int sel = 0;
					sel = SendMessage(mod->ListhWnd, LB_GETCURSEL ,0,0);
					mod->RemoveNode(sel);
					break;
					}


				}
			break;


		}
	return FALSE;
	}






//--- Affect region mod methods -------------------------------

AFRMod::AFRMod() 
	{
	MakeRefByID(
		FOREVER, PBLOCK_REF, 
		CreateParameterBlock(
			descVer0, PBLOCK_LENGTH, CURRENT_VERSION));
	
	pblock->SetValue(PB_SIZE,0,30.0f);
	pblock->SetValue(PB_TENSION,0,1.0f);
	pblock->SetValue(PB_RENDER,0,3.0f);
	pblock->SetValue(PB_VIEW,0,6.0f);

	iBut = NULL;
	SelfNode = NULL;
	OtherNode.ZeroCount();
	}

void AFRMod::BeginEditParams(
		IObjParam  *ip, ULONG flags,Animatable *prev)
	{
	this->ip = ip;

	
	TimeValue t = ip->GetTime();
	NotifyDependents(Interval(t,t), PART_ALL, REFMSG_BEGIN_EDIT);
	NotifyDependents(Interval(t,t), PART_ALL, REFMSG_MOD_DISPLAY_ON);
	SetAFlag(A_MOD_BEING_EDITED);

	pmapParam = CreateCPParamMap(
		descParam,PARAMDESC_LENGTH,
		pblock,
		ip,
		hInstance,
		MAKEINTRESOURCE(IDD_AFRPARAM),
		GetString(IDS_RB_PARAMETERS),
		0);	
	pmapParam->SetUserDlgProc(new MapDlgProc(this));

	}

void AFRMod::EndEditParams(
		IObjParam *ip,ULONG flags,Animatable *next)
	{
	this->ip = NULL;
	TimeValue t = ip->GetTime();

	// NOTE: This flag must be cleared before sending the REFMSG_END_EDIT
	ClearAFlag(A_MOD_BEING_EDITED);

	NotifyDependents(Interval(t,t), PART_ALL, REFMSG_END_EDIT);
	NotifyDependents(Interval(t,t), PART_ALL, REFMSG_MOD_DISPLAY_OFF);


	DestroyCPParamMap(pmapParam);
	}

RefTargetHandle AFRMod::Clone(RemapDir& remap)
	{
	AFRMod *mod = new AFRMod();
	mod->ReplaceReference(PBLOCK_REF,pblock->Clone(remap));
	return mod;
	}



int AFRMod::Display(
		TimeValue t, INode* inode, ViewExp *vpt, 
		int flagst, ModContext *mc)
	{

	return 0;

	}



int AFRMod::RenderBegin(TimeValue t, ULONG flags)
	{
	SetAFlag(A_RENDER);
//	ivalid.SetEmpty();
	NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
	return 0;
	}

int AFRMod::RenderEnd(TimeValue t)
	{
	ClearAFlag(A_RENDER);
//	ivalid.SetEmpty();
	NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);
	return 0;

	}




class MyEnumProc : public DependentEnumProc 
	{
      public :
      virtual int proc(ReferenceMaker *rmaker); 
      INodeTab Nodes;              
	};

int MyEnumProc::proc(ReferenceMaker *rmaker) 
	{ 
	if (rmaker->SuperClassID()==BASENODE_CLASS_ID)    
			{
            Nodes.Append(1, (INode **)&rmaker);                 
			}
     return 0;              
	}


void AFRMod::ModifyObject(
		TimeValue t, ModContext &mc, ObjectState *os, INode *node)
	{	
	if (SelfNode == NULL)
		{
	     MyEnumProc dep;              
		 EnumDependents(&dep);
		 SelfNode = dep.Nodes[0];

		}
	Matrix3 BaseTM(1);
	if (SelfNode != NULL)
		 BaseTM = Inverse(SelfNode->GetObjectTM(t));
//	Interval iv = FOREVER;
	Interval iv = os->obj->ChannelValidity (t, GEOM_CHAN_NUM);
	iv &= os->obj->ChannelValidity (t, TOPO_CHAN_NUM);
	iv &= os->obj->ChannelValidity (t, SELECT_CHAN_NUM);
	iv &= os->obj->ChannelValidity (t, SUBSEL_TYPE_CHAN_NUM);
	iv &= os->obj->ChannelValidity (t, TEXMAP_CHAN_NUM);
	iv &= os->obj->ChannelValidity (t, VERT_COLOR_CHAN_NUM);

	for (int nc =0; nc < OtherNode.Count(); nc++)
		{
		if (OtherNode[nc] != NULL) 
			{
			OtherNode[nc]->GetObjTMBeforeWSM(t,&iv);
			ObjectState osp = OtherNode[nc]->EvalWorldState (t);
			iv &= osp.obj->ObjectValidity (t);

			}

		}

	pblock->GetValue(PB_SIZE,t,size,iv);
	pblock->GetValue(PB_TENSION,t,tension,iv);
	pblock->GetValue(PB_RENDER,t,coursenessR,iv);
	pblock->GetValue(PB_VIEW,t,coursenessV,iv);
	pblock->GetValue(PB_AUTO,t,autoCourse,iv);
	pblock->GetValue(PB_HIDE,t,hideEdges,iv);

	float c;

	if (TestAFlag(A_RENDER))
		c = coursenessR;
	else c = coursenessV;

	if (autoCourse)
		{
		c = size/c ;
		}

//get point list

	Point3 *plist;


	int count = os->obj->NumPoints();

	for (int i = 0; i < OtherNode.Count(); i++)
		{



		ObjectState tos =  OtherNode[i]->EvalWorldState(t,TRUE);		
		if (tos.obj->GetInterface(I_PARTICLEOBJ)!=NULL)
			{
			SimpleParticle *particles;
			ObjectState pos = OtherNode[i]->EvalWorldState(t);
			particles = (SimpleParticle*) pos.obj;
			particles->Update(t,OtherNode[i]);

			for (int j=0; j<particles->parts.Count(); j++) {
				if (particles->parts.ages[j] >= 0) 
					count++;
				}

			}
		else if (tos.obj->SuperClassID()==GEOMOBJECT_CLASS_ID)
			{

//			count += tos.obj->NumPoints();
			BOOL needDel;
			NullView nullView;
			Mesh *mesh = ((GeomObject*)tos.obj)->GetRenderMesh(t,OtherNode[i],nullView,needDel);
			if (mesh)
				{
				count += mesh->getNumVerts();
				}
			if (needDel) delete mesh;

			}

		}


	plist = new Point3[count];
	int ct = 0;
	for (i=0; i<os->obj->NumPoints(); i++) 
		{
		plist[ct++] = os->obj->GetPoint(i);
		}

	for (i = 0; i < OtherNode.Count(); i++)
		{

		Matrix3 tm = OtherNode[i]->GetObjectTM(t);
		tm = tm * BaseTM ;
		ObjectState tos =  OtherNode[i]->EvalWorldState(t,TRUE);		
		if (tos.obj->GetInterface(I_PARTICLEOBJ)!=NULL)
			{
			SimpleParticle *particles;
			ObjectState pos = OtherNode[i]->EvalWorldState(t);
			particles = (SimpleParticle*) pos.obj;
			particles->Update(t,OtherNode[i]);

			for (int j=0; j<particles->parts.Count(); j++) {
				if (particles->parts.ages[j] >= 0) 
					{
					plist[ct++] = particles->parts[j] * BaseTM;
					}
				}	

			}
		else if (tos.obj->SuperClassID()==GEOMOBJECT_CLASS_ID)
			{
//			int vcount = tos.obj->NumPoints();
//			for (int j = 0; j < vcount; j++)
//				{
//				plist[ct++] = tos.obj->GetPoint(j)*tm;
//				}


			BOOL needDel;
			NullView nullView;
			Mesh *mesh = ((GeomObject*)tos.obj)->GetRenderMesh(t,OtherNode[i],nullView,needDel);
			if (mesh)
				{
				int vcount = mesh->getNumVerts();
				for (int j = 0; j < vcount; j++)
					{
					plist[ct++] = mesh->getVert(j) *tm;
					}
				}
	
			if (needDel) delete mesh;

			}


		}




	TriObject *tri = CreateNewTriObject();
	float thres=0.6f;
	blobP.CreateBlobs(plist,count,&tri->mesh,thres,c,tension,size);	

/*	if (hideEdges)
		{
		for(i=0;tri->mesh.numFaces;i++)
			{
			tri->mesh.faces[i].setEdgeVisFlags(EDGE_INVIS,EDGE_INVIS,EDGE_INVIS);
			}
		}
*/	

//delete plist

	delete [] plist;

	tri->SetChannelValidity(TOPO_CHAN_NUM, iv);
	tri->SetChannelValidity(GEOM_CHAN_NUM, iv);
	tri->SetChannelValidity(TEXMAP_CHAN_NUM, iv);
	tri->SetChannelValidity(MTL_CHAN_NUM, iv);
	tri->SetChannelValidity(SELECT_CHAN_NUM, iv);
	tri->SetChannelValidity(SUBSEL_TYPE_CHAN_NUM, iv);
	tri->SetChannelValidity(DISP_ATTRIB_CHAN_NUM, iv);

	os->obj = tri;
	os->obj->UnlockObject();

//	os->obj->UpdateValidity(GEOM_CHAN_NUM,iv);		

	}

Interval AFRMod::LocalValidity(TimeValue t)
	{
	Interval iv = FOREVER;
	float v;
	Point3 pt;
	pblock->GetValue(PB_SIZE,t,v,iv);
	pblock->GetValue(PB_TENSION,t,v,iv);
	pblock->GetValue(PB_RENDER,t,v,iv);
	pblock->GetValue(PB_VIEW,t,v,iv);

	for (int nc =0; nc < OtherNode.Count(); nc++)
		{
		if (OtherNode[nc] != NULL) 
			{
			OtherNode[nc]->GetObjTMBeforeWSM(t,&iv);
			ObjectState osp = OtherNode[nc]->EvalWorldState (t);
			iv &= osp.obj->ObjectValidity (t);

			}

		}
	

	return iv;
	}


void AFRMod::RemoveNode(int sel)

{

if (sel>=0)
	{
	SendMessage(ListhWnd, LB_DELETESTRING  ,(WPARAM) sel,0);

	for (int j=sel;j<(OtherNode.Count()-1);j++)
		{
		ReplaceReference(j+1,OtherNode[j+1]);
		}

	DeleteReference(OtherNode.Count());

	OtherNode.Delete(OtherNode.Count()-1,1);

	NotifyDependents(FOREVER, TOPO_CHANNEL, REFMSG_CHANGE);
	}


}




RefTargetHandle AFRMod::GetReference(int i)
	{
	switch (i) {
		case PBLOCK_REF: return (RefTargetHandle)pblock;
		default : 
			{
			if ( (i >=1) && (i - 1) < OtherNode.Count())
				return (RefTargetHandle)OtherNode[i-1];
			}
		}
	return NULL;
	}

void AFRMod::SetReference(int i, RefTargetHandle rtarg)
	{
	switch (i) {
		case PBLOCK_REF: pblock = (IParamBlock*)rtarg; break;
		default:
			{
			if ( (i >=1) && (i - 1) < OtherNode.Count())
				OtherNode[i-1] = (INode*)rtarg; break;
			}
		}

	}


Animatable* AFRMod::SubAnim(int i)
	{
	switch (i) {
		case PBLOCK_REF: 		return pblock;
		default: 			return NULL;   
		}
	}

TSTR AFRMod::SubAnimName(int i)
	{

		 return _T(""); 
		
	}

RefResult AFRMod::NotifyRefChanged(
		Interval changeInt,RefTargetHandle hTarget, 
		PartID& partID, RefMessage message)
	{
	switch (message) {
		case REFMSG_GET_PARAM_DIM: {
			GetParamDim *gpd = (GetParamDim*)partID;
			switch (gpd->index) {
				case PB_SIZE: gpd->dim = stdWorldDim; break;
				}
			return REF_STOP; 
			}

		case REFMSG_GET_PARAM_NAME: {
			GetParamName *gpn = (GetParamName*)partID;			
			switch (gpn->index) {
				case PB_SIZE: gpn->name = GetString(IDS_SIZE); break;
				}
			return REF_STOP; 
			}

		case REFMSG_CHANGE:	
			{ 
//			  if (pmapParam) pmapParam->Invalidate();
			break;
			}

		case REFMSG_TARGET_DELETED:	
			{ if (hTarget==SelfNode) SelfNode=NULL;
			  for (int i = 0; i < OtherNode.Count(); i++)
				{
				  if (hTarget==OtherNode[i]) 
					{
					  RemoveNode(i);
					}
				}

			break;
			}


		}
	return REF_SUCCEED;
	}






IOResult AFRMod::Load(ILoad *iload)
	{
	Modifier::Load(iload);
	IOResult res = IO_OK;
	ULONG nb;

	OtherNode.ZeroCount();
	while (IO_OK==(res=iload->OpenChunk())) {
		int id = iload->CurChunkID();
		switch(id)  {

			case NODE_COUNT_CHUNK: 
				{
				int c;
				iload->Read(&c,sizeof(c),&nb);
				OtherNode.SetCount(c);
				for (int i=0; i<c; i++)  
					{
					OtherNode[i] = NULL;
					}
				break;
				}
			}
		iload->CloseChunk();
		if (res!=IO_OK) 
			return res;
		}	


//build selection list now
//	int s;
	
	return IO_OK;
	}

IOResult AFRMod::Save(ISave *isave)
	{
	Modifier::Save(isave);
	ULONG nb;


	int c = OtherNode.Count();
	isave->BeginChunk(NODE_COUNT_CHUNK);
	isave->Write(&c,sizeof(c),&nb);
	isave->EndChunk();



	
	return IO_OK;
	}

