#include "max.h"
#include "particle.h"
#include "metapart.h"

// Fix 2.1
//  AddToSet: correct plst list handling and a puzzle
//  AddABlob: use indirect values to get right particle
//  CalcField: have to use indirect addressing of particles
// BLOBSETS: calculate blob sub-sets Y/N
//  if not used, only calculation is excluded, code remains in file.
//#define BLOBSETS

// HAB7FACES: there can be more than 6 faces ...
#define HAB7FACES 

// Speedup:
// complexity is NXxNYxNZxParticles
// HABSPEED: speed factor 20+
// - don't use PartcleSystem as parameter
// - don't use Tab,it tends to be slow
// HABSPEED2: speed factor 5+
// - don't resize mesh to often it is slow, use blocked resize
//
// HABLOOKLIKE: speed factor 1+-
// - your particles are smaller!
//
#define HABLOOKLIKE
// HABASSERT: bug catch helper
//
#undef HABASSERT
// HABBUGFIX:used for a bugfix (see there)
//
#define HABBUGFIX

#ifndef __METAPARTS__H

#define TOPF 0
#define FRONTF 1
#define RIGHTF 2	
#define BACKF 3	
#define LEFTF 4	
#define BOTF 5	
typedef struct{
 Point3 vmin,vmax;
 int pstart,pstop;
}BunchData;
typedef struct{
  int pattern,v[4];
}linedata;
typedef struct
{ float field;
  linedata oldfaces[3];
} corners;
typedef struct{
 Point3 pos;
 int gridnum;
 float field;
 BOOL sign;
}cubepts;
typedef struct { 
int p1,p2,e1a,e1b,e2a,e2b;
}bseg;

class BlobParticle {
	public:
		Tab<BunchData> bloblst;

		Tab<int> plst;
		Tab<Point3> maxpoint;
		Tab<Point3> minpoint;
		Tab<float> rsquare;
		float tens;

		float fthreshold;
		int bnum,cblob,Nx,Ny,Nz;
		Point3 deltax,deltay,deltaz,deltaxy;
		corners *level0,*level1;
		int vertex,face;
		void RecalcBounds(int num,Point3 pvmin,Point3 pvmax);
		BOOL CheckBounds(int num,Point3 pvmin,Point3 pvmax);
		void AddABlob(int c,int *incount);
		BOOL OneInSet(int bnum) {return(bloblst[bnum].pstart==bloblst[bnum].pstop);}
		void AddToSet(int bcount,int going);
		void MergeSet(int bcount);
		void FillInFields(corners *level,int start,int ix,int y,cubepts *cube,Point3 *arts);
		float CalcField(Point3 cpos,Point3 *arts);
		Point3 CheckForVertex(Point3 apos,Point3 bpos,float afield,float bfield);
		int FindVNum(int edge,linedata oldrec);
		int GetOldVertex(int oldface,int newface,linedata oldrec);
		int CheckCurBlst(bseg *blst,int bcount,int a, int b);
		void AddLine(bseg *blst,int *bcount,int *vlst,int a,int b,int *sign,int *edge);
		void FillInFace(cubepts *cube,bseg *blst,int *bcount,int a,int b,int c, int d,linedata *srec,int *oldv,linedata *olddata,Mesh *mesh,Point3 *arts);
		void FillInCube(cubepts *cube,bseg *blst,int *bcount,Mesh *mesh,Point3 *arts);
		void GenerateFaces(bseg *line,int lcount,Mesh *mesh);
		int AllocateCubes(int levelmax);
		void DoCube(int ix,int x,int y,int z,Point3 cpos,Point3 *arts,Mesh *mesh);
//		int CreateBlobs(Point3 *arts,Mesh *mesh,float threshold,float res,float strength, float size);
		int CreateBlobs(Point3 *parts,int count, Mesh *mesh,float threshold,float res,float strength,float size);
	};
#endif

void BlobParticle::RecalcBounds(int num,Point3 pvmin,Point3 pvmax)
{ BOOL inset=FALSE;
  if (pvmin.x<bloblst[num].vmin.x)
		bloblst[num].vmin.x=pvmin.x;
  if (pvmax.x>bloblst[num].vmax.x) 
		bloblst[num].vmax.x=pvmax.x;
  if (pvmin.y<bloblst[num].vmin.y) 
		bloblst[num].vmin.y=pvmin.y;
  if (pvmax.y>bloblst[num].vmax.y)
		bloblst[num].vmax.y=pvmax.y;
  if (pvmin.z<bloblst[num].vmin.z)
		bloblst[num].vmin.z=pvmin.z;
  if (pvmax.z>bloblst[num].vmax.z) 
		bloblst[num].vmax.z=pvmax.z;
}

void BlobParticle::AddToSet(int bcount,int going)
{ int *plst2,i,pos;
	// FIX 2.1 another puzzle ...
//  if (bloblst[bcount].pstop=bloblst[going].pstart-1)
  if (bloblst[bcount].pstop==bloblst[going].pstart-1)
	bloblst[bcount].pstop=bloblst[going].pstop;
  else
  { int len=bloblst[going].pstop-(pos=bloblst[going].pstart)+1;
    int endlst=bloblst[bcount].pstop+1;
    plst2=new int[len];
	// FIX 2.1: we have to fill plst2 ...
	for (i=0;i<len;i++)
//	  plst2[0]=plst[pos++];
	  plst2[i]=plst[pos++];
	pos=bloblst[going].pstart-1;
	// FIX 2.1: we have to go thru plst ...
//	for (i=plst[bloblst[going].pstop];i>plst[bloblst[bcount].pstart];i--)
//	  plst[i]=plst[pos--];
//	pos=bloblst[bcount].pstop+1;
	for (i=bloblst[going].pstop;i>bloblst[bcount].pstart;i--)
	  plst[i]=plst[pos--];
	pos=bloblst[bcount].pstart;
	// FIX 2.1: we have to fill plst ...
	for (i=0;i<len;i++)
		plst[pos++]=plst2[i];
//		plst[pos]=plst2[i];
	bloblst[bcount].pstop+=len;
	delete[] plst2;
	// Fix 2.1: we have to move the bloblst indeces of the moved blobs
	assert(bcount <= going);
	for (i=bcount+1; i <going;i++) {
		bloblst[i].pstart+=len;
		bloblst[i].pstop+=len;
	}
  }
  RecalcBounds(bcount,bloblst[going].vmin,bloblst[going].vmax);
  bloblst.Delete(going,1);
  bnum--;
}

BOOL BlobParticle::CheckBounds(int num,Point3 pvmin,Point3 pvmax)
{ return(((bloblst[num].vmin.x<=pvmin.x && pvmin.x<=bloblst[num].vmax.x)||(bloblst[num].vmin.x<=pvmax.x && pvmax.x<=bloblst[num].vmax.x))&&
		 ((bloblst[num].vmin.y<=pvmin.y && pvmin.y<=bloblst[num].vmax.y)||(bloblst[num].vmin.y<=pvmax.y && pvmax.y<=bloblst[num].vmax.y))&&
		 ((bloblst[num].vmin.z<=pvmin.z && pvmin.z<=bloblst[num].vmax.z)||(bloblst[num].vmin.z<=pvmax.z && pvmax.z<=bloblst[num].vmax.z)));
}

void BlobParticle::MergeSet(int bcount)
{ int curblob=0;			
  BOOL inset;
  while (curblob<bnum)
  { inset=FALSE;
    curblob=0;
	while ((curblob<=bnum)&&((curblob==bcount)||(!(inset=CheckBounds(bcount,bloblst[curblob].vmin,bloblst[curblob].vmax)))))
      curblob++;
    if (inset)
    { if (curblob<bcount) {AddToSet(curblob,bcount);bcount=curblob;}
      else 
	   AddToSet(bcount,curblob);
	}
  }
}

void BlobParticle::AddABlob(int c,int *incount)
{ int i;
  Point3 tmin,tmax;
  BOOL inset=FALSE;
  // Fix 2.1: have to use indirect addressing, as plst changes
  int pIn=plst[*incount];
//  assert(plst[*incount]==*incount);
//  bloblst[bnum].vmin=minpoint[*incount];
//  bloblst[bnum].vmax=maxpoint[*incount];
  bloblst[bnum].vmin=minpoint[pIn];
  bloblst[bnum].vmax=maxpoint[pIn];
  int tmp,bcount=0;
  // if particle fits into an existing set, add it
//  while ((bcount<bnum)&&(OneInSet(bcount)||!(inset=CheckBounds(bcount,minpoint[*incount],maxpoint[*incount]))))
  while ((bcount<bnum)&&(OneInSet(bcount)||!(inset=CheckBounds(bcount,minpoint[pIn],maxpoint[pIn]))))
   bcount++;
  (*incount)++;
  if (inset)   // see if new set not intersects any other set
  { AddToSet(bcount,bnum);
    MergeSet(bcount);
  }
  else 	// create new set by checking all unset particles with this one
  { for (i=*incount;i<c;i++)
    { int j=plst[i];
//	  assert(i==plst[i]);
//	  inset=CheckBounds(bnum,minpoint[i],maxpoint[i]);
	  inset=CheckBounds(bnum,minpoint[j],maxpoint[j]);
	  if (inset)
	  {
//		RecalcBounds(bnum,minpoint[i],maxpoint[i]);
		RecalcBounds(bnum,minpoint[j],maxpoint[j]);
		if (*incount!=i)
	    { tmp=plst[i];
	      plst[i]=plst[*incount];
	      plst[*incount]=tmp;
	    }
	    (*incount)++;
	  }
    }
    bloblst[bnum].pstop=*incount-1;
	if (!OneInSet(bnum)) MergeSet(bnum);
  }
}

int BlobParticle::AllocateCubes(int levelmax)
{  BOOL ok;
   ok=(NULL!=(level0=new corners[levelmax]))&&
	(NULL!=(level1=new corners[levelmax]));
   if (ok) 
	for (int i=0;i<levelmax;i++)
	{ level0[i].oldfaces[0].pattern=0;
	  level0[i].oldfaces[1].pattern=0;
	  level0[i].oldfaces[2].pattern=0;
	  level1[i].oldfaces[0].pattern=0;
	  level1[i].oldfaces[1].pattern=0;
	  level1[i].oldfaces[2].pattern=0;
    }
   return(ok);
}

#ifdef HABSPEED
float BlobParticle::CalcField(Point3 cpos)
#else
float BlobParticle::CalcField(Point3 cpos,Point3 *arts)
#endif
{ float d2,field=0.0f;
#ifndef HABSPEED
	float temp;
#endif
  Point3 dist;
  for (int i=bloblst[cblob].pstart;i<=bloblst[cblob].pstop;i++)
#ifndef HABSPEED
  { dist=cpos-parts[plst[i]];
    d2=dist.x*dist.x+dist.y*dist.y+dist.z*dist.z;
	if (d2<rsquare[i]) 
	{  temp=(1-d2/rsquare[i]);
//	   field+=parts.tension[plst[i]]*temp*temp;
	   field+=tens*temp*temp;
	}
  }
#else
  { // FIX 2.1: plst order can change, 
	// so use indirect addressing
	int c=plst[i];
	dist=cpos-pospoint[c];
    d2=dist.x*dist.x+dist.y*dist.y+dist.z*dist.z;
	d2-=rsquare[c];
	if (d2<0.0f) 
	   field+=tensionOverR4[c]*d2*d2;
  }
#endif
  return(field-fthreshold);
}

#ifdef HABSPEED

// HABBUGFIX is used here
// took me nearly 2 hours to catch this!
// Are you a fan of Feuerbach,C puzzle? or PC-lint
// puzzle solved wrong, right side of assignment is evaluated 1st!
// does PC-Lint find that ?
// I read Feuerbach, so in general I'm not in need of lint programs.
// ---------------------------------------------------------------
void BlobParticle::FillInFields(corners *level,int start,int ix,int y,cubepts *cube)
{ int s=start;
  if (ix==0)
  { if (y==0) level[cube[start].gridnum].field=CalcField(cube[start].pos);	
#ifdef HABBUGFIX
    s=start+3;
    level[cube[s].gridnum].field=CalcField(cube[s].pos);
#else
    level[cube[s=start+3].gridnum].field=CalcField(cube[s].pos);
#endif
  }
  if (y==0) {
#ifdef HABBUGFIX
	  s=start+1;
    level[cube[s].gridnum].field=CalcField(cube[s].pos);
#else
    level[cube[s=start+1].gridnum].field=CalcField(cube[s].pos);
#endif
  }
#ifdef HABBUGFIX
  s=start+2;
  level[cube[s].gridnum].field=CalcField(cube[s].pos); 
#else
  level[cube[s=start+2].gridnum].field=CalcField(cube[s].pos); 
#endif
  for (int i=start;i<start+4;i++)
    cube[i].sign=((cube[i].field=level[cube[i].gridnum].field)>=0.0f);
}

#else

void BlobParticle::FillInFields(corners *level,int start,int ix,int y,cubepts *cube,Point3 *arts)
{ int s=start;
  if (ix==0)
  { if (y==0) level[cube[start].gridnum].field=CalcField(cube[start].pos,parts);	
    level[cube[s=start+3].gridnum].field=CalcField(cube[s].pos,parts);
  }
  if (y==0) 
    level[cube[s=start+1].gridnum].field=CalcField(cube[s].pos,parts);
  level[cube[s=start+2].gridnum].field=CalcField(cube[s].pos,parts); 
  for (int i=start;i<start+4;i++)
    cube[i].sign=((cube[i].field=level[cube[i].gridnum].field)>=0.0f);
}

#endif

Point3 BlobParticle::CheckForVertex(Point3 apos,Point3 bpos,float afield,float bfield)
{ Point3 v;
  if (afield>bfield)
  	v=apos-(apos-bpos)*(afield/(afield-bfield));
  else v=bpos-(bpos-apos)*(bfield/(bfield-afield));
  return(v);
}

int BlobParticle::FindVNum(int edge,linedata oldrec)
{// I'm not checking for no vertex.If it doesn't exist, the program's really screwed up.
  int vnum;
  if (edge==0) {if (oldrec.pattern>8) vnum=oldrec.v[1];else vnum=oldrec.v[0];}
  else if (edge==1)
  { if ((oldrec.pattern==1)||(oldrec.pattern==7)||(oldrec.pattern==12)||(oldrec.pattern==13))
     vnum=oldrec.v[1];
    else if ((oldrec.pattern==4)||(oldrec.pattern==5)||(oldrec.pattern==9)||(oldrec.pattern==16))
	  vnum=oldrec.v[0];
	else if (oldrec.pattern==15) vnum=oldrec.v[2];
	else vnum=oldrec.v[3];
  }
  else if (edge==2)
  {	if ((oldrec.pattern==2)||(oldrec.pattern==4)||(oldrec.pattern==14))
      vnum=oldrec.v[1];
    else if ((oldrec.pattern==6)||(oldrec.pattern==10)||(oldrec.pattern==12))
	  vnum=oldrec.v[0];
    else if ((oldrec.pattern==7)||(oldrec.pattern==8))
	  vnum=oldrec.v[2];
	else vnum=oldrec.v[3];
  }
  else if (edge==3)
  {	if ((oldrec.pattern==3)||(oldrec.pattern==5)||(oldrec.pattern==6)||(oldrec.pattern==8))
      vnum=oldrec.v[1];
    else if ((oldrec.pattern==11)||(oldrec.pattern==13)||(oldrec.pattern==14)||(oldrec.pattern==15))
	  vnum=oldrec.v[0];
    else if (oldrec.pattern==16)
	  vnum=oldrec.v[2];
	else vnum=oldrec.v[3];
  }
  return(vnum);
}
int BlobParticle::GetOldVertex(int oldface,int newface,linedata oldrec)
{ int edge;
  if (oldface==TOPF)
  {	if (newface==FRONTF) edge=0;
    else if (newface==RIGHTF) edge=1;
	else if (newface==BACKF) edge=2;
	else if (newface==LEFTF) edge=3;
  } 
  else if (oldface==FRONTF)
  {	if (newface==TOPF) edge=2;
    else if (newface==RIGHTF) edge=1;
	else if (newface==BOTF) edge=0;
	else if (newface==LEFTF) edge=3;
  } 
  else if (oldface==2)	//HAB:RIGHTF
  {	if (newface==FRONTF) edge=1;
    else if (newface==TOPF) edge=2;
	else if (newface==BACKF) edge=3;
	else if (newface==BOTF) edge=0;
  }
  return(FindVNum(edge,oldrec));
}
int BlobParticle::CheckCurBlst(bseg *blst,int bcount,int a, int b)
{ int oldnum=-1;
  for (int i=0;i<bcount;i++)
  {if (((blst[i].e1a==b)&&(blst[i].e1b==a))||((blst[i].e1a==a)&&(blst[i].e1b==b)))
   { oldnum=blst[i].p1;
	break;
    }
	if (((blst[i].e2a==b)&&(blst[i].e2b==a))||((blst[i].e2a==a)&&(blst[i].e2b==b)))
   { oldnum=blst[i].p2;
	break;
    }
  }
return (oldnum);
}

void BlobParticle::AddLine(bseg *blst,int *bcount,int *vlst,int a,int b,int *sign,int *edge)
{ blst[*bcount].p1=vlst[a];
  blst[*bcount].e1a=sign[a];
  blst[*bcount].e1b=edge[a];
  blst[*bcount].e2a=sign[b];
  blst[*bcount].e2b=edge[b];
  blst[(*bcount)++].p2=vlst[b];
}

#ifdef HABSPEED
void BlobParticle::FillInFace(cubepts *cube,bseg *blst,int *bcount,int a,int b,int c, int d,linedata *srec,int *oldv,linedata *olddata,Mesh *mesh)
#else
void BlobParticle::FillInFace(cubepts *cube,bseg *blst,int *bcount,int a,int b,int c, int d,linedata *srec,int *oldv,linedata *olddata,Mesh *mesh,Point3 *arts)
#endif
{int sign[4],center,pat=0,mvcount=0,oldind=1,vlst[4],vcount=0,oldnum,edge[4],obcount=*bcount;
 Point3 mainvarr[12],midpos;
 BOOL save=(srec!=NULL); 
    if (cube[a].sign!=cube[b].sign)
	  { sign[vcount]=a;edge[vcount]=b;
	    if (!oldv[oldind]&&(oldv[3]==TOPF))
		  vlst[vcount++]=GetOldVertex(oldind,oldv[3],olddata[oldind]);
		else if ((oldnum=CheckCurBlst(blst,*bcount,a,b))>-1)
		{ vlst[vcount++]=oldnum;}
		else
		{ vlst[vcount++]=vertex+mvcount;
	      mainvarr[mvcount++]=CheckForVertex(cube[a].pos,cube[b].pos,cube[a].field,cube[b].field);
		}
	  }
    if (cube[b].sign!=cube[c].sign)
	  { sign[vcount]=b;edge[vcount]=c;
	    if ((!oldv[oldind]&&(oldv[3]==LEFTF))||(!oldv[oldind=2]&&(oldv[3]==BACKF)))
		  vlst[vcount++]=GetOldVertex(oldind,oldv[3],olddata[oldind]);
		else if ((oldnum=CheckCurBlst(blst,*bcount,b,c))>-1)
		{ vlst[vcount++]=oldnum;}
		else
		{ vlst[vcount++]=vertex+mvcount;
	      mainvarr[mvcount++]=CheckForVertex(cube[b].pos,cube[c].pos,cube[b].field,cube[c].field);
		}
	  }
    if (cube[c].sign!=cube[d].sign)
	  { sign[vcount]=c;edge[vcount]=d;
	    if ((!oldv[oldind=0]&&(oldv[3]!=BOTF))||(!oldv[oldind=1]&&(oldv[3]==BOTF)))
		  vlst[vcount++]=GetOldVertex(oldind,oldv[3],olddata[oldind]);
		else if ((oldnum=CheckCurBlst(blst,*bcount,c,d))>-1)
		{ vlst[vcount++]=oldnum;}
		else
		{ vlst[vcount++]=vertex+mvcount;
	      mainvarr[mvcount++]=CheckForVertex(cube[c].pos,cube[d].pos,cube[c].field,cube[d].field);
		}
	  }
    if ((vcount==1)||(vcount==3))
	  { sign[vcount]=d;edge[vcount]=a;
	    if ((!oldv[oldind=2]&&((oldv[3]!=RIGHTF)&&(oldv[3]!=BACKF)))||(!oldv[oldind=1]&&(oldv[3]==RIGHTF)))
		  vlst[vcount++]=GetOldVertex(oldind,oldv[3],olddata[oldind]);
		else if ((oldnum=CheckCurBlst(blst,*bcount,d,a))>-1)
		{ vlst[vcount++]=oldnum;}
		else
		{ vlst[vcount++]=vertex+mvcount;
	      mainvarr[mvcount++]=CheckForVertex(cube[d].pos,cube[a].pos,cube[d].field,cube[a].field);
		}
	  }
	if (vcount>0)
	{
// pattern values are reversed because we'll be using them on the next cube.
	BOOL top=(oldv[3]==BOTF);
	if (vcount==2) 
	{ BOOL rev=FALSE;
      if (cube[sign[0]].sign>0)  //right order within the cube
	  {	AddLine(blst,bcount,vlst,1,0,sign,edge);rev=TRUE;}
	  else {AddLine(blst,bcount,vlst,0,1,sign,edge);}
	  if (save)
	  {	if (sign[0]==c) 
	      pat=(rev?(top?3:12):(top?11:4));
	    else if (sign[0]==b) 
	    { if (sign[1]==c) pat=(rev?(top?9:14):(top?1:6)); 
	      else pat=(rev?(top?5:13):(top?13:5));
	    }
	    else if (sign[1]==b) 
		  pat=(rev?(top?12:3):(top?4:11));
	    else if 
		  (sign[1]==c) pat=(rev?(top?10:2):(top?2:10));
	    else 
		  pat=(rev?(top?6:1):(top?14:9));
	    (*srec).pattern=pat;
	    (*srec).v[0]=blst[obcount].p2;
	    (*srec).v[1]=blst[obcount].p1;
	  }
	}
	else
	{ midpos=(cube[a].pos+cube[c].pos)/2.0f; 
#ifdef HABSPEED
	  center=(CalcField(midpos)>=0.0f);
#else
	  center=(CalcField(midpos,parts)>=0.0f);
#endif
	  if ((!center)&&(!cube[sign[0]].sign)) 
	  { AddLine(blst,bcount,vlst,0,1,sign,edge);
		AddLine(blst,bcount,vlst,2,3,sign,edge);
		pat=15;}
	  else if (center)
	  {	if (cube[sign[0]].sign)
		{AddLine(blst,bcount,vlst,1,0,sign,edge);
		 AddLine(blst,bcount,vlst,3,2,sign,edge); 
		 pat=8;}
	    else
		{ AddLine(blst,bcount,vlst,0,3,sign,edge);
	      AddLine(blst,bcount,vlst,2,1,sign,edge); 
		  pat=16;}
	  }														
	  else
	  { AddLine(blst,bcount,vlst,3,0,sign,edge);
	    AddLine(blst,bcount,vlst,1,2,sign,edge);
		pat=7;}
	  if (save) 
	  { int inobc=obcount+1;
		(*srec).pattern=pat;
		 if (top)
		 { (*srec).v[0]=blst[inobc].p2;
		   (*srec).v[1]=blst[inobc].p1;
		   (*srec).v[2]=blst[obcount].p2;
		   (*srec).v[3]=blst[obcount].p1;
		 }
		 else
		 { (*srec).v[0]=blst[obcount].p2;
		   (*srec).v[1]=blst[obcount].p1;
		   (*srec).v[2]=blst[inobc].p2;
		   (*srec).v[3]=blst[inobc].p1;
		 }
  	  }
	}
	if (mvcount>0)
#ifdef HABSPEED2
	{
		if(mvcount>vFree) {
			vAlloc+=vBlock;
			mesh->setNumVerts(vAlloc,TRUE);
			vFree+=vBlock;
		}
	  for (int i=0;i<mvcount;i++)
	   mesh->verts[vertex++]=mainvarr[i];
	  vFree-=mvcount;
	}
#else
	{ mesh->setNumVerts(vertex+mvcount,TRUE);
	  for (int i=0;i<mvcount;i++)
	   mesh->verts[vertex++]=mainvarr[i];
	}
#endif
  }
  else 	if (save) (*srec).pattern=17;
}

void GetSavedFace(bseg *lineseg,int *lcount,linedata srec)
{ int i,j=0,vsaved=0;
   if (srec.pattern!=17)
   { vsaved=((srec.pattern<7)||((srec.pattern>8)&&(srec.pattern<15))?1:2);
	 for (i=0;i<vsaved;i++)
	 { lineseg[*lcount].p1=srec.v[j++];
	   lineseg[*lcount].e1a=0;
	   lineseg[*lcount].e1b=0;
	   lineseg[*lcount].e2a=0;
	   lineseg[*lcount].e2b=0;
	   lineseg[*lcount].p2=srec.v[j++];
	   (*lcount)++;
	 }
   }
}

#ifdef HABSPEED
void BlobParticle::FillInCube(cubepts *cube,bseg *blst,int *bcount,Mesh *mesh)
{ int i;
  linedata *olddata=level0[cube[0].gridnum].oldfaces;
  int old[4];
  for (i=0;i<3;i++) old[i]=(olddata[i].pattern==0);
  if (old[0])
  { old[3]=TOPF;FillInFace(cube,blst,bcount,0,1,2,3,NULL,old,olddata,mesh);}
  else GetSavedFace(blst,bcount,level0[cube[0].gridnum].oldfaces[0]);
  // front
  if (old[1])
  { old[3]=FRONTF;FillInFace(cube,blst,bcount,5,6,1,0,NULL,old,olddata,mesh);}
  else GetSavedFace(blst,bcount,level0[cube[0].gridnum].oldfaces[1]);
  // right
   old[3]=RIGHTF;FillInFace(cube,blst,bcount,6,7,2,1,&level0[cube[1].gridnum].oldfaces[2],old,olddata,mesh);
  // back
   old[3]=BACKF;FillInFace(cube,blst,bcount,7,8,3,2,&level0[cube[3].gridnum].oldfaces[1],old,olddata,mesh);
  //left
  if (old[2])
  { old[3]=LEFTF;FillInFace(cube,blst,bcount,8,5,0,3,NULL,old,olddata,mesh);}
  else GetSavedFace(blst,bcount,level0[cube[0].gridnum].oldfaces[2]);
  //bottom
  old[3]=BOTF;FillInFace(cube,blst,bcount,8,7,6,5,&level1[cube[5].gridnum].oldfaces[0],old,olddata,mesh);
}

#else
void BlobParticle::FillInCube(cubepts *cube,bseg *blst,int *bcount,Mesh *mesh,Point3 *arts)
{ int i;
  linedata *olddata=level0[cube[0].gridnum].oldfaces;
  int old[4];
  for (i=0;i<3;i++) old[i]=(olddata[i].pattern==0);
  if (old[0])
  { old[3]=TOPF;FillInFace(cube,blst,bcount,0,1,2,3,NULL,old,olddata,mesh,parts);}
  else GetSavedFace(blst,bcount,level0[cube[0].gridnum].oldfaces[0]);
  // front
  if (old[1])
  { old[3]=FRONTF;FillInFace(cube,blst,bcount,5,6,1,0,NULL,old,olddata,mesh,parts);}
  else GetSavedFace(blst,bcount,level0[cube[0].gridnum].oldfaces[1]);
  // right
   old[3]=RIGHTF;FillInFace(cube,blst,bcount,6,7,2,1,&level0[cube[1].gridnum].oldfaces[2],old,olddata,mesh,parts);
  // back
   old[3]=BACKF;FillInFace(cube,blst,bcount,7,8,3,2,&level0[cube[3].gridnum].oldfaces[1],old,olddata,mesh,parts);
  //left
  if (old[2])
  { old[3]=LEFTF;FillInFace(cube,blst,bcount,8,5,0,3,NULL,old,olddata,mesh,parts);}
  else GetSavedFace(blst,bcount,level0[cube[0].gridnum].oldfaces[2]);
  //bottom
  old[3]=BOTF;FillInFace(cube,blst,bcount,8,7,6,5,&level1[cube[5].gridnum].oldfaces[0],old,olddata,mesh,parts);
}

#endif
void BlobParticle::AddFace(int a, int b, int c,int face,Mesh *pm,int curmtl)
{
#ifdef HABASSERT
	if(!(a>=0&&a<vertex))
		assert(a>=0&&a<vertex);
	if(!(b>=0&&b<vertex))
		assert(b>=0&&b<vertex);
	if(!(c>=0&&c<vertex))
		assert(c>=0&&c<vertex);
#endif
  pm->faces[face].setSmGroup(1);
  pm->faces[face].setVerts(a,b,c);
  pm->faces[face].setMatID((MtlID)curmtl);
  pm->faces[face].setEdgeVisFlags(1,1,0);

}

void BlobParticle::GenerateFaces(bseg *line,int lcount,Mesh *mesh)
{ bseg temp;
  int vlst[12],vcount=0,start=0;
  for (int i=0;i<lcount;i++)
  {	for (int j=i+1;j<lcount;j++)
	{ if (line[j].p1==line[i].p2)
	  {	temp=line[j];
		line[j]=line[i+1];
		line[i+1]=temp;
		break;
	  }
	}
  }
  i=0;
  while (i<lcount) 
  { vlst[0]=line[i].p1;vlst[1]=line[i++].p2;
    vcount=2;
    start=vlst[0];
    while (line[i].p2!=start)
  	 vlst[vcount++]=line[i++].p2;
	i++;
#ifdef HABSPEED2
	if((vcount-2)>faceFree){
		faceAlloc+=faceBlock;
		mesh->setNumFaces(faceAlloc,TRUE);
		faceFree+=faceBlock;
	}
// somethings going wrong here!
//	fFree-=(vcount-2);
#else
    mesh->setNumFaces(face+vcount-2,TRUE);
#endif

    AddFace(vlst[0],vlst[1],vlst[2],face++,mesh,0);
#ifdef HABSPEED2
	faceFree--;
#endif
    if (vcount>3)
    { if (vcount<6) 
      { AddFace(vlst[0],vlst[2],vlst[3],face++,mesh,0);
        if (vcount==5) AddFace(vlst[0],vlst[3],vlst[4],face++,mesh,0);
#ifdef HABSPEED2
		faceFree-=(1+(vcount==5));
#endif
      }
	// Fix 2.1 : handle 6 points faces
      else if(vcount==6)
	  { AddFace(vlst[2],vlst[3],vlst[4],face++,mesh,0);
	    AddFace(vlst[4],vlst[5],vlst[0],face++,mesh,0);
	    AddFace(vlst[0],vlst[2],vlst[4],face++,mesh,0);
#ifdef HABSPEED
		faceFree-=3;
#endif
	  } // end else
	  // Fix 2.1: can be more than 6 points ..
	  else {
		  for (int iface=3; iface<vcount;iface++)
		    AddFace(vlst[0],vlst[iface-1],vlst[iface],face++,mesh,7);
		faceFree-=(vcount-3);
	  }
    }  // end if
  }	 // end while
}

#ifdef HABSPEED
void BlobParticle::DoCube(int ix,int x,int y,int z,Point3 cpos,Mesh *mesh)
#else
void BlobParticle::DoCube(int ix,int x,int y,int z,Point3 cpos,Point3 *arts,Mesh *mesh)
#endif
{ int rowmax=Nx,otherside=x+rowmax,lcount=0;
  bseg line[12];
  Point3 bpos=cpos-deltaz;
  cubepts curcube[10];
  curcube[0].pos=cpos;curcube[0].gridnum=x;
  curcube[1].pos=cpos+deltax;curcube[1].gridnum=x+1;
  curcube[2].pos=cpos+deltaxy;curcube[2].gridnum=otherside+1;
  curcube[3].pos=cpos+deltay;curcube[3].gridnum=otherside;
  curcube[4].pos=cpos;curcube[4].gridnum=x;
  curcube[5].pos=bpos;curcube[5].gridnum=curcube[0].gridnum;
  curcube[6].pos=bpos+deltax;curcube[6].gridnum=curcube[1].gridnum;
  curcube[7].pos=bpos+deltaxy;curcube[7].gridnum=curcube[2].gridnum;
  curcube[8].pos=bpos+deltay;curcube[8].gridnum=curcube[3].gridnum;
  curcube[9].pos=bpos;curcube[9].gridnum=curcube[0].gridnum;
#ifdef HABSPEED
  if (z==0)	FillInFields(level0,0,ix,y,curcube);
#else
  if (z==0)	FillInFields(level0,0,ix,y,curcube,parts);
#endif
  else for (int i=0;i<4;i++)
    curcube[i].sign=((curcube[i].field=level0[curcube[i].gridnum].field)>=0.0f);
#ifdef HABSPEED
  FillInFields(level1,5,ix,y,curcube);
  FillInCube(curcube,line,&lcount,mesh);
#else
  FillInFields(level1,5,ix,y,curcube,parts);
  FillInCube(curcube,line,&lcount,mesh,parts);
#endif
  if (lcount>0) GenerateFaces(line,lcount,mesh);
}

//int BlobParticle::CreateBlobs(Point3 *arts,Mesh *mesh,float threshold,float res,float strength, float size)
int BlobParticle::CreateBlobs(Point3 *parts,int count, Mesh *mesh,float threshold,float res,float strength,float size)
{ int i,ix,x,y,z,incount,levelmax;
  int c=0,total=count;

  Point3 deltas,cpos;

  tens = strength;

  fthreshold=threshold;
  for (i=0; i<total; i++)
    c++;
  if (c>0)
  {
  plst.SetCount(c);
  maxpoint.SetCount(c);
  minpoint.SetCount(c);
#ifdef HABSPEED
  rsquare = new float[c];
  tensionOverR4 = new float[c];
  pospoint = new Point3[c];
#else
  rsquare.SetCount(c);
#endif
  c=0;
  float maxRadius=0.0f;
  for (i=0; i<total; i++)
	if (1)//(parts.Alive(i))  
	{ plst[c]=i;
#ifdef HABSPEED
//watje
	  float radius=size;//parts.radius[i];
	  pospoint[c]=parts[i];
#ifdef HABLOOKLIKE
	  radius /= 2.0f;
	  if(radius>maxRadius)
		  maxRadius=radius;
	  minpoint[c]=pospoint[c]-radius;
	  maxpoint[c]=pospoint[c]+radius;
	  rsquare[c]=radius*radius;
#endif
//	  tensionOverR4[c]=parts.tension[i]/(rsquare[c]*rsquare[c]);
	  tensionOverR4[c]=tens/(rsquare[c]*rsquare[c]);
#else
	  minpoint[c]=parts[plst[c]]-size;//parts.radius[plst[c]];
	  maxpoint[c]=parts[plst[c]]+size;//parts.radius[plst[c]];
	  rsquare[c]=size*size;//parts.radius[plst[c]]*parts.radius[plst[c]];
//	  minpoint[c]=parts[plst[c]]-parts.radius[plst[c]];
//	  maxpoint[c]=parts[plst[c]]+parts.radius[plst[c]];
//	  rsquare[c]=parts.radius[plst[c]]*parts.radius[plst[c]];
#endif
      c++;
	}
#ifdef BLOBSETS
  bnum=0;
  incount=0;
  bloblst.SetCount(c);
  while (incount<c)
  {	bloblst[bnum].pstop=(bloblst[bnum].pstart=incount);
    AddABlob(c,&incount);
	bnum++;
  }
#else
  bloblst.SetCount(1);
  bloblst[0].pstart=0;
  bloblst[0].pstop=c-1;
  bloblst[0].vmax=maxpoint[0];
  bloblst[0].vmin=minpoint[0];
  bnum=1;
  for (incount=1; incount<c;incount++)
	  RecalcBounds(0, minpoint[incount], maxpoint[incount]);
#endif
  vertex=0;face=0;
#ifdef HABSPEED2
  vAlloc=vFree=0;vBlock=4096;
  faceAlloc=faceFree=0;faceBlock=1024;
#endif
  corners *temp;
  maxpoint.ZeroCount();maxpoint.Shrink();
  minpoint.ZeroCount();minpoint.Shrink();
  int maxx,maxy;
  // Fix 2.1: need this no more to looklike ...
  // now it's independent of fudgy value, as it should be
//float fudgy=maxRadius*strength;
  float fudgy=0.5f*res*strength;

  Point3 fudgept=Point3(fudgy,fudgy,fudgy);
  BOOL finishedok=FALSE;
  for (i=0;i<bnum;i++)
  { cblob=i;
	deltas=(bloblst[i].vmax+=fudgept)-(bloblst[i].vmin-=fudgept);
	Nx=int(1+deltas.x/res);
	Ny=int(1+deltas.y/res);
	Nz=int(1+deltas.z/res);
	maxx=Nx-1;maxy=Ny-1;
	deltas.x/=Nx;deltas.y/=Ny;deltas.z/=Nz;
	deltax=Point3(deltas.x,0.0f,0.0f);deltay=Point3(0.0f,deltas.y,0.0f);
	deltaz=Point3(0.0f,0.0f,deltas.z);deltaxy=Point3(deltas.x,deltas.y,0.0f);
	levelmax=Nx*Ny;
	if (!AllocateCubes(levelmax)) return(0);
	cpos=bloblst[i].vmin;
	cpos.z=bloblst[i].vmax.z;
	x=0;
	for (z=0;z<Nz;z++)
	{ for (y=0;y<maxy;y++)
      { for (ix=0;ix<maxx;ix++)
		{ 
#ifdef HABSPEED
		  DoCube(ix,x,y,z,cpos,mesh);
#else
		  DoCube(ix,x,y,z,cpos,parts,mesh);
#endif
		  cpos.x+=deltas.x;
		  x++;
		if (GetAsyncKeyState (VK_ESCAPE)) goto arrgh;
		}
	    x++;
	    cpos.x=bloblst[i].vmin.x;
		cpos.y+=deltas.y;
	  }
	  x=0;
	  temp=level0;
      level0=level1;
      level1=temp;
	  cpos.y=bloblst[i].vmin.y;
	  cpos.z-=deltas.z;
    }
  delete[] level0;
  delete[] level1;
  }	
  finishedok=TRUE;
arrgh:
  if (!finishedok) 
  { delete[] level0;
	delete[] level1;
    mesh->setNumVerts(0);
	mesh->setNumFaces(0);
  }
#ifdef HABSPEED2
	else {
		mesh->setNumVerts(vertex,TRUE);
		mesh->setNumFaces(face,TRUE);
	}
#endif

  mesh->setNumTVerts(0);
  mesh->setNumTVFaces(0);
  bloblst.ZeroCount();bloblst.Shrink();
  plst.ZeroCount();plst.Shrink();
#ifdef HABSPEED
  delete [] rsquare;
  delete [] tensionOverR4;
  delete [] pospoint;
#else
  rsquare.ZeroCount();rsquare.Shrink();
#endif
  }
  else
  { mesh->setNumVerts(0);
    mesh->setNumFaces(0);
	mesh->setNumTVerts(0);
	mesh->setNumTVFaces(0);
  }
  return 1;
}	

/*
int MetaParticle::CreateBlobs(ParticleSys parts,Mesh *mesh,float threshold,float res,float strength)
{ BlobParticle blob;

  return(blob.CreateBlobs(parts,mesh,threshold,res,strength));
}
*/